
#include <iostream>
using std::cout;
using std::cin;
using std::cerr;
using std::endl;

#include <string>
using std::string;

#include "RefCount.h"

#include "ResolveNameToObject.h"

#include <malloc.h>

#define DoThis(_ex)                                                  \
   if (CORBA::_ex::_narrow(const_cast<CORBA::Exception *>(&e)))      \
   {                                                                 \
      return "CORBA::" #_ex;                                         \
   }                                                                 \

static std::string GetNameFromCORBAException(const CORBA::Exception &e)
{
   OMNIORB_FOR_EACH_SYS_EXCEPTION(DoThis)

   if (CORBA::UserException::_narrow(const_cast<CORBA::Exception *>(&e))) 
   { 
      return "CORBA::UserException";
   }

   return "Unrecognised CORBA::Exception!";
}

class heapInfo
{
   public :

      heapInfo() 
         : m_used(0), m_free(0), m_blocks(0)
      {
         _HEAPINFO heapInfo;   
         int heapStatus;

         heapInfo._pentry = NULL;         // start at the begining...

         while ( _HEAPOK == (heapStatus = _heapwalk(&heapInfo)))
         {
            if (heapInfo._useflag == _USEDENTRY)
            {
               m_used += heapInfo._size;
            }
            else
            {
               m_free += heapInfo._size;
            }

            m_blocks++;
         }

         if (heapStatus != _HEAPEMPTY &&
             heapStatus != _HEAPEND)
         {
            cout << "Heap error" << endl;
         }
      }

      unsigned long used() const { return m_used; }
      unsigned long free() const { return m_free; }
      unsigned long blocks() const { return m_blocks; }

   private :

      unsigned long m_used;
      unsigned long m_free;
      unsigned long m_blocks;
};


int main (int argc, char **argv) 
{
   try 
   {
	
      CORBA::ORB_var orb = CORBA::ORB_init(argc, argv, "omniORB3");

      CORBA::Object_var obj = ResolveNameToObject(orb, "JetByte Test", "NamedCounterFactory");

      NamedCounterFactory_var counterFactory = NamedCounterFactory::_narrow(obj);

      if (!CORBA::is_nil(counterFactory))
      {
         bool ok = false;

         if (argc > 1)
         {
            if (argc == 2)
            {
               if (0 == stricmp("ping", argv[1]))
               {
                  cout << "List of named counters:" << endl;

                  counterSeq_var seq;

                  NamedCounterIterator_var it;

                  counterFactory->GetIterator(0, seq, it);

                  if (!CORBA::is_nil(it))
                  {
                     const size_t timeout = it->GetTimeout();

                     cout << "Timeout = " << timeout << " seconds" << endl;

                     for (int i = 0; i < 5; ++i)
                     {
                        it->Ping();

                        cout << "Ping" << endl;

                        const size_t sleepTime = timeout - 1;

                        cout << "Sleep for " << sleepTime << " seconds" << endl;

                        ::Sleep(1000 * sleepTime);
                     }

                     cout << "Sleep for " << timeout + 1 << " seconds" << endl;

                     ::Sleep(1000 * (timeout + 1));
                     
                     cout << "Attempt a ping" << endl;

                     it->Ping();

                     cout << "Ping" << endl;


                  }

                  cout << "Done." << endl;


                  ok = true;
               }
               else if (0 == stricmp("list", argv[1]))
               {
                  heapInfo before;

                  counterSeq_var seq = counterFactory->GetAll();

                  cout << "List of named counters:" << endl;

                  for (size_t i = 0; i < seq->length(); ++i)
                  {
                     cout << seq[i]->name() << " - " << seq[i]->value() << endl;

                     seq[i]->Release();
                  }

                  cout << "Done." << endl;

                  heapInfo after;

                  cout << "Heap usage:" << endl;
                  cout << "Before: " << endl;
                  cout << "Used   : " << before.used() << endl;
                  cout << "Free   : " << before.free() << endl;
                  cout << "Total  : " << before.used() + before.free() << endl;
                  cout << "Blocks : " << before.blocks() << endl;

                  cout << "After: " << endl;
                  cout << "Used   : " << after.used() << endl;
                  cout << "Free   : " << after.free() << endl;
                  cout << "Total  : " << after.used() + before.free() << endl;
                  cout << "Blocks : " << after.blocks() << endl;

                  cout << "Difference: " << endl;
                  cout << "Used   : " << after.used() - before.used() << endl;
                  cout << "Free   : " << after.free() - before.free()<< endl;
                  cout << "Total  : " << (after.used() + before.free()) - (after.used() + before.free())<< endl;
                  cout << "Blocks : " << after.blocks() - before.blocks() << endl;

                  ok = true;
               }
               else if (0 == stricmp("listro", argv[1]))
               {
                  heapInfo before;

                  counterDataSeq_var seq = counterFactory->GetAllReadOnly();

                  cout << "List of named counters:" << endl;

                  for (size_t i = 0; i < seq->length(); ++i)
                  {
                     cout << (char*)seq[i].name << " - " << seq[i].value << endl;
                  }

                  cout << "Done." << endl;

                  heapInfo after;

                  cout << "Heap usage:" << endl;
                  cout << "Before: " << endl;
                  cout << "Used   : " << before.used() << endl;
                  cout << "Free   : " << before.free() << endl;
                  cout << "Total  : " << before.used() + before.free() << endl;
                  cout << "Blocks : " << before.blocks() << endl;

                  cout << "After: " << endl;
                  cout << "Used   : " << after.used() << endl;
                  cout << "Free   : " << after.free() << endl;
                  cout << "Total  : " << after.used() + before.free() << endl;
                  cout << "Blocks : " << after.blocks() << endl;

                  cout << "Difference: " << endl;
                  cout << "Used   : " << after.used() - before.used() << endl;
                  cout << "Free   : " << after.free() - before.free()<< endl;
                  cout << "Total  : " << (after.used() + before.free()) - (after.used() + before.free())<< endl;
                  cout << "Blocks : " << after.blocks() - before.blocks() << endl;

                  ok = true;
               }
 
            }
            else if (argc == 3)
            {
               if (0 == stricmp("add", argv[1]))
               {
                  counterFactory->AddNamedCounter(argv[2], 0);

                  cout << "Counter " << argv[2] << " added with a value of 0" << endl;
                  
                  ok = true;
               }
               else if (0 == stricmp("del", argv[1]))
               {
                  counterFactory->RemoveNamedCounter(argv[2]);

                  cout << "Counter " << argv[2] << " deleted" << endl;
   
                  ok = true;
               }
               else if (0 == stricmp("enum", argv[1]))
               {
                  heapInfo before;

                  cout << "List of named counters:" << endl;

                  EnumNamedCounter_var enumRef = counterFactory->GetEnumerator();

                  counterSeq_var seq = new counterSeq();

                  CORBA::ULong numCounters;

                  const size_t c_countersPerCall = atoi(argv[2]);

                  do
                  {
	                  numCounters = enumRef->Next(c_countersPerCall, seq);

	                  for (size_t i = 0; i < numCounters; i++)
	                  {
                        cout << seq[i]->name() << " - " << seq[i]->value() << endl;

                        seq[i]->Release();
                     }
                     cout << "--" << endl;
                  }
                  while (numCounters == c_countersPerCall );

                  enumRef->Release();

                  cout << "Done." << endl;

                  ok = true;

                  heapInfo after;

                  cout << "Heap usage:" << endl;
                  cout << "Before: " << endl;
                  cout << "Used   : " << before.used() << endl;
                  cout << "Free   : " << before.free() << endl;
                  cout << "Total  : " << before.used() + before.free() << endl;
                  cout << "Blocks : " << before.blocks() << endl;

                  cout << "After: " << endl;
                  cout << "Used   : " << after.used() << endl;
                  cout << "Free   : " << after.free() << endl;
                  cout << "Total  : " << after.used() + before.free() << endl;
                  cout << "Blocks : " << after.blocks() << endl;

                  cout << "Difference: " << endl;
                  cout << "Used   : " << after.used() - before.used() << endl;
                  cout << "Free   : " << after.free() - before.free()<< endl;
                  cout << "Total  : " << (after.used() + before.free()) - (after.used() + before.free())<< endl;
                  cout << "Blocks : " << after.blocks() - before.blocks() << endl;
               }
               else if (0 == stricmp("enumro", argv[1]))
               {
                  heapInfo before;

                  cout << "List of named counters:" << endl;

                  EnumNamedCounterData_var enumRef = counterFactory->GetEnumeratorReadOnly();

                  counterDataSeq_var seq = new counterDataSeq();

                  CORBA::ULong numCounters;

                  const size_t c_countersPerCall = atoi(argv[2]);

                  do
                  {
	                  numCounters = enumRef->Next(c_countersPerCall, seq);

	                  for (size_t i = 0; i < numCounters; i++)
	                  {
                        cout << (char*)seq[i].name << " - " << seq[i].value << endl;
                     }
                     cout << "--" << endl;
                  }
                  while (numCounters == c_countersPerCall );

                  enumRef->Release();

                  cout << "Done." << endl;

                  ok = true;

                  heapInfo after;

                  cout << "Heap usage:" << endl;
                  cout << "Before: " << endl;
                  cout << "Used   : " << before.used() << endl;
                  cout << "Free   : " << before.free() << endl;
                  cout << "Total  : " << before.used() + before.free() << endl;
                  cout << "Blocks : " << before.blocks() << endl;

                  cout << "After: " << endl;
                  cout << "Used   : " << after.used() << endl;
                  cout << "Free   : " << after.free() << endl;
                  cout << "Total  : " << after.used() + before.free() << endl;
                  cout << "Blocks : " << after.blocks() << endl;

                  cout << "Difference: " << endl;
                  cout << "Used   : " << after.used() - before.used() << endl;
                  cout << "Free   : " << after.free() - before.free()<< endl;
                  cout << "Total  : " << (after.used() + before.free()) - (after.used() + before.free())<< endl;
                  cout << "Blocks : " << after.blocks() - before.blocks() << endl;
               }
            }
            else if (argc == 4)
            {
               if (0 == stricmp("change", argv[1]))
               {
                  NamedCounter_var counter = counterFactory->GetNamedCounter(argv[2]);
               
                  const long initialVal = counter->value();
                  const long newVal = initialVal + atoi(argv[3]);

                  counter->value(newVal);

                  counter->Release();

                  cout << "Counter " << argv[2] << " initial value = " << initialVal << " new value = " << newVal << endl;

                  ok = true;
               }
               else if (0 == stricmp("hold", argv[1]))
               {
                  NamedCounter_var counter = counterFactory->GetNamedCounter(argv[2]);
  
                  cout << "Counter " << argv[2] << " held" << endl;
   
                  Sleep(1000*atoi(argv[3]));

                  long value = counter->value();

                  counter->Release();

                  cout << "Counter " << argv[2] << " value = " << value << endl;

                  ok = true;
               }
               else if (0 == stricmp("holdit", argv[1]))
               {
                  counterSeq_var seq;

                  NamedCounterIterator_var it;

                  counterFactory->GetIterator(atoi(argv[2]), seq, it);

                  cout << "Iterator held" << endl;

                  Sleep(1000*atoi(argv[3]));

                  cout << "Process Iterator" << endl;

                  // handle initial batch

                  size_t i = 0;

                  while (i < seq->length())
                  {
                     cout << seq[i]->name() << " - " << seq[i]->value() << endl;

                     seq[i]->Release();
                  
                     i++;
                  }

                  cout << "--" << endl;

                  // handle subsequent batches...

                  if (!CORBA::is_nil(it))
                  {
                     bool moreCounters = false;

                     do
                     {
                        moreCounters = it->next_n(atoi(argv[3]), seq);

                        if (moreCounters)
                        {
                           i = 0;

                           while (i < seq->length())
                           {
                              cout << seq[i]->name() << " - " << seq[i]->value() << endl;

                              seq[i]->Release();
                     
                              i++;
                           }
                           cout << "--" << endl;
                        }
                     }
                     while (moreCounters);

                     //it->destroy();
                  }

                  cout << "Done." << endl;

                  ok = true;
               }
               else if (0 == stricmp("enum2", argv[1]))
               {
                  cout << "List of named counters:" << endl;

                  EnumNamedCounter_var enumRef = counterFactory->GetEnumerator();

                  counterSeq_var seq = new counterSeq();

                  CORBA::ULong numCounters;

                  const size_t c_countersPerCall = atoi(argv[2]);

                  do
                  {
	                  numCounters = enumRef->Next(c_countersPerCall, seq);

	                  for (size_t i = 0; i < numCounters; i++)
	                  {
                        cout << seq[i]->name() << " - " << seq[i]->value() << endl;

                        seq[i]->Release();
                     }
                     cout << "--" << endl;

                     if (numCounters == c_countersPerCall)
                     {
                        ::Sleep(1000 * atoi(argv[3]));
                     }
                  }
                  while (numCounters == c_countersPerCall );

                  enumRef->Release();

                  cout << "Done." << endl;

                  ok = true;
               }
               else if (0 == stricmp("it", argv[1]))
               {
                  heapInfo before;

                  cout << "List of named counters:" << endl;

                  // number to get in first call, 
                  // number to get in subsequent calls,

                  counterSeq_var seq;

                  NamedCounterIterator_var it;

                  counterFactory->GetIterator(atoi(argv[2]), seq, it);

                  // handle initial batch

                  size_t i = 0;

                  while (i < seq->length())
                  {
                     cout << seq[i]->name() << " - " << seq[i]->value() << endl;

                     seq[i]->Release();
                  
                     i++;
                  }
               
                  cout << "--" << endl;

                  // handle subsequent batches...

                  if (!CORBA::is_nil(it))
                  {
                     bool moreCounters = false;

                     do
                     {
                        moreCounters = it->next_n(atoi(argv[3]), seq);

                        if (moreCounters)
                        {
                           i = 0;

                           while (i < seq->length())
                           {
                              cout << seq[i]->name() << " - " << seq[i]->value() << endl;

                              seq[i]->Release();
                     
                              i++;
                           }
                           cout << "--" << endl;
                        }
                     }
                     while (moreCounters);

                     it->destroy();
                  }

                  cout << "Done." << endl;

                  ok = true;

                  heapInfo after;

                  cout << "Heap usage:" << endl;
                  cout << "Before: " << endl;
                  cout << "Used   : " << before.used() << endl;
                  cout << "Free   : " << before.free() << endl;
                  cout << "Total  : " << before.used() + before.free() << endl;
                  cout << "Blocks : " << before.blocks() << endl;

                  cout << "After: " << endl;
                  cout << "Used   : " << after.used() << endl;
                  cout << "Free   : " << after.free() << endl;
                  cout << "Total  : " << after.used() + before.free() << endl;
                  cout << "Blocks : " << after.blocks() << endl;

                  cout << "Difference: " << endl;
                  cout << "Used   : " << after.used() - before.used() << endl;
                  cout << "Free   : " << after.free() - before.free()<< endl;
                  cout << "Total  : " << (after.used() + before.free()) - (after.used() + before.free())<< endl;
                  cout << "Blocks : " << after.blocks() - before.blocks() << endl;
               }
               else if (0 == stricmp("itx", argv[1]))
               {
                  heapInfo before;

                  cout << "List of named counters:" << endl;

                  // number to get in first call, 
                  // number to get in subsequent calls,

                  counterSeq_var seq;

                  NamedCounterIterator_var it;

                  counterFactory->GetIterator(atoi(argv[2]), seq, it);

                  // handle initial batch

                  size_t i = 0;

                  while (i < seq->length())
                  {
                     cout << seq[i]->name() << " - " << seq[i]->value() << endl;

                     seq[i]->Release();
                  
                     i++;
                  }
               
                  cout << "--" << endl;

                  // handle subsequent batches...

                  if (!CORBA::is_nil(it))
                  {
                     bool moreCounters = false;

                     do
                     {
                        moreCounters = it->next_n(atoi(argv[3]), seq);

                        if (moreCounters)
                        {
                           i = 0;

                           while (i < seq->length())
                           {
                              cout << seq[i]->name() << " - " << seq[i]->value() << endl;

                              seq[i]->Release();
                     
                              i++;
                           }
                           cout << "--" << endl;
                        }
                     }
                     while (moreCounters);

                     //it->destroy();
                  }

                  cout << "Done." << endl;

                  ok = true;

                  heapInfo after;

                  cout << "Heap usage:" << endl;
                  cout << "Before: " << endl;
                  cout << "Used   : " << before.used() << endl;
                  cout << "Free   : " << before.free() << endl;
                  cout << "Total  : " << before.used() + before.free() << endl;
                  cout << "Blocks : " << before.blocks() << endl;

                  cout << "After: " << endl;
                  cout << "Used   : " << after.used() << endl;
                  cout << "Free   : " << after.free() << endl;
                  cout << "Total  : " << after.used() + before.free() << endl;
                  cout << "Blocks : " << after.blocks() << endl;

                  cout << "Difference: " << endl;
                  cout << "Used   : " << after.used() - before.used() << endl;
                  cout << "Free   : " << after.free() - before.free()<< endl;
                  cout << "Total  : " << (after.used() + before.free()) - (after.used() + before.free())<< endl;
                  cout << "Blocks : " << after.blocks() - before.blocks() << endl;
               }
               else if (0 == stricmp("itro", argv[1]))
               {
                  heapInfo before;

                  cout << "List of named counters:" << endl;

                  // number to get in first call, 
                  // number to get in subsequent calls,

                  counterDataSeq_var seq;

                  NamedCounterDataIterator_var it;

                  counterFactory->GetIteratorReadOnly(atoi(argv[2]), seq, it);

                  // handle initial batch

                  size_t i = 0;

                  while (i < seq->length())
                  {
                     cout << (char*)seq[i].name << " - " << seq[i].value << endl;

                     i++;
                  }
               
                  cout << "--" << endl;

                  // handle subsequent batches...

                  if (!CORBA::is_nil(it))
                  {
                     bool moreCounters = false;

                     do
                     {
                        moreCounters = it->next_n(atoi(argv[3]), seq);

                        if (moreCounters)
                        {
                           i = 0;

                           while (i < seq->length())
                           {
                              cout << (char*)seq[i].name << " - " << seq[i].value << endl;

                              i++;
                           }
                           cout << "--" << endl;
                        }
                     }
                     while (moreCounters);

                     it->destroy();
                  }

                  cout << "Done." << endl;

                  ok = true;

                  heapInfo after;

                  cout << "Heap usage:" << endl;
                  cout << "Before: " << endl;
                  cout << "Used   : " << before.used() << endl;
                  cout << "Free   : " << before.free() << endl;
                  cout << "Total  : " << before.used() + before.free() << endl;
                  cout << "Blocks : " << before.blocks() << endl;

                  cout << "After: " << endl;
                  cout << "Used   : " << after.used() << endl;
                  cout << "Free   : " << after.free() << endl;
                  cout << "Total  : " << after.used() + before.free() << endl;
                  cout << "Blocks : " << after.blocks() << endl;

                  cout << "Difference: " << endl;
                  cout << "Used   : " << after.used() - before.used() << endl;
                  cout << "Free   : " << after.free() - before.free()<< endl;
                  cout << "Total  : " << (after.used() + before.free()) - (after.used() + before.free())<< endl;
                  cout << "Blocks : " << after.blocks() - before.blocks() << endl;
               }
 
            }
         }

         if (!ok)
         {
            cout << "usage: client list          - list all counters and values" << endl;
            cout << "       client listro        - list all counter data" << endl;
            cout << "       client add <counter> - add a new counter" << endl;
            cout << "       client del <counter> - remove a counter" << endl;
            cout << "       client change <counter> <value> - change the counter by value" << endl;
            cout << "       client hold <counter> <secs> - obtain the counter and hold onto it for" << endl;
            cout << "                                      secs seconds" << endl;
            cout << "       client enum <max>    - enumerate all counters, max at a time" << endl;
            cout << "       client enumro <max>  - enumerate all counter data, max at a time" << endl;
            cout << "       client it <f> <n>    - iterates all counters, f on the first call" << endl;
            cout << "                              and n on subsequent calls" << endl;
            cout << "       client itro <f> <n>  - iterates all counter data, f on the first call" << endl;
            cout << "                              and n on subsequent calls" << endl;
         }
      }
      else
      {
         cout << "Failed to narrow counter factory" << endl;
      }

      orb->destroy();
   }
   catch (NamedCounterFactory::CounterExists &e)
   {
      cerr << "Counter \"" << (char*)e.name << "\" already exists!" << endl;
   }
   catch (NamedCounterFactory::NoCounterExists &e)
   {
      cerr << "Counter \"" << (char*)e.name << "\" doesn't exist!" << endl;
   }
   catch(CORBA::Exception &e) 
   {
      cerr << "Caught CORBA::Exception: " << GetNameFromCORBAException(e) << endl;
   }
   catch(omniORB::fatalException& fe) 
   {
      cerr << "Caught omniORB::fatalException:" << endl;
      cerr << "  file: " << fe.file() << endl;
      cerr << "  line: " << fe.line() << endl;
      cerr << "  mesg: " << fe.errmsg() << endl;
   }
   catch(...) 
   {
      cerr << "Caught unknown exception." << endl;
   }

   return 0;
}


